from datetime import datetime, timedelta
import json
import sys
from typing import (
    Any,
    Generic,
    IO,
    NoReturn,
    Optional,
    TypeVar,
    Union,
    overload,
)


def format_timedelta(delta: timedelta) -> str:
    values = [
        (delta.days, 'd'),
        (delta.seconds, 's'),
        (delta.microseconds, 'us'),
    ]
    values = ['%d%s' % v for v in values if v[0]]
    if values:
        return ' '.join(values)
    else:
        return '0s'


class JSONDateEncoder(json.JSONEncoder):
    def default(self, obj: Union[timedelta, datetime, object]) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            return format_timedelta(obj)
        return super().default(obj)


def open_or_stdin(filename: str, stdin: IO[str] = sys.stdin) -> IO[str]:
    if filename == '-':
        fo = stdin
    else:
        fo = open(filename)
    return fo


T = TypeVar("T")


class PassthroughManager(Generic[T]):
    def __init__(self, ret: T) -> None:
        self.ret = ret

    def __enter__(self) -> T:
        return self.ret

    def __exit__(self, *a: Any) -> None:
        pass


@overload
def open_or_return(fo_or_path: None, mode: str = 'r') -> NoReturn:
    ...


@overload
def open_or_return(fo_or_path: str, mode: str = 'r') -> IO[str]:
    ...


@overload
def open_or_return(
    fo_or_path: IO[str], mode: str = 'r'
) -> PassthroughManager[IO[str]]:
    ...


def open_or_return(
    fo_or_path: Optional[Union[str, IO[str]]], mode: str = 'r'
) -> Union[IO[str], PassthroughManager[IO[str]]]:
    # Returns a context manager around a file-object for fo_or_path. If
    # fo_or_path is a file-object, the context manager keeps it open. If it's a
    # path, the file is opened with mode and will be closed upon context exit.
    # If fo_or_path is None, a ValueError is raised.

    if fo_or_path is None:
        raise ValueError('No file-like object nor path provided')
    if isinstance(fo_or_path, str):
        return open(fo_or_path, mode)

    # Skip default file context manager. This allows to always use with
    # statement and don't care about closing the file. If the file is opened
    # here, it will be closed properly. Otherwise, it will be kept open thanks
    # to PassthroughManager.
    return PassthroughManager(fo_or_path)


class Timer:
    def __enter__(self) -> "Timer":
        self.start = datetime.utcnow()
        return self

    def __exit__(self, *a: Any) -> None:
        self.delta = datetime.utcnow() - self.start



""".. currentmodule:: pgtoolkit.hba

This module supports reading, validating, editing and rendering ``pg_hba.conf``
file. See `Client Authentication
<https://www.postgresql.org/docs/current/static/auth-pg-hba-conf.html>`__ in
PostgreSQL documentation for details on format and values of ``pg_hba.conf``
file.


API Reference
-------------

The main entrypoint of this API is the :func:`parse` function. It returns a
:class:`HBA` object containing :class:`HBARecord` instances.

.. autofunction:: parse
.. autoclass:: HBA
.. autoclass:: HBARecord


Examples
--------

Loading a ``pg_hba.conf`` file :

.. code:: python

    pgpass = parse('my_pg_hba.conf')

You can also pass a file-object:

.. code:: python

    with open('my_pg_hba.conf', 'r') as fo:
        hba = parse(fo)

Creating a ``pg_hba.conf`` file from scratch :

.. code:: python

    hba = HBA()
    record = HBARecord(
        conntype='local', database='all', user='all', method='peer',
    )
    hba.lines.append(record)

    with open('pg_hba.conf', 'w') as fo:
        hba.write(fo)


Using as a script
-----------------

:mod:`pgtoolkit.hba` is usable as a CLI script. It accepts a pg_hba file path
as first argument, read it, validate it and re-render it. Fields are aligned to
fit pseudo-column width. If filename is ``-``, stdin is read instead.

.. code:: console

    $ python -m pgtoolkit.hba - < data/pg_hba.conf
    # TYPE  DATABASE        USER            ADDRESS                 METHOD

    # "local" is for Unix domain socket connections only
    local   all             all                                     trust
    # IPv4 local connections:
    host    all             all             127.0.0.1/32            ident map=omicron

"""  # noqa

import os
import shlex
import sys
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import warnings

from .errors import ParseError
from ._helpers import (
    open_or_return,
    open_or_stdin,
)


class HBAComment(str):
    def __repr__(self) -> str:
        return '<%s %.32s>' % (self.__class__.__name__, self)


class HBARecord:
    """Holds a HBA record composed of fields and a comment.

    Common fields are accessible through attribute : ``conntype``,
    ``databases``, ``users``, ``address``, ``netmask``, ``method``.
    Auth-options fields are also accessible through attribute like ``map``,
    ``ldapserver``, etc.

    ``address`` and ``netmask`` fields are not always defined. If not,
    accessing undefined attributes trigger an :exc:`AttributeError`.

    ``databases`` and ``users`` have a single value variant respectively
    :attr:`database` and :attr:`user`, computed after the list representation
    of the field.

    .. automethod:: parse
    .. automethod:: __init__
    .. automethod:: __str__
    .. automethod:: matches
    .. autoattribute:: database
    .. autoattribute:: user

    """

    COMMON_FIELDS = [
        'conntype', 'databases', 'users', 'address', 'netmask', 'method',
    ]
    CONNECTION_TYPES = ['local', 'host', 'hostssl', 'hostnossl']

    @classmethod
    def parse(cls, line: str) -> "HBARecord":
        """Parse a HBA record

        :rtype: :class:`HBARecord` or a :class:`str` for a comment or blank
                line.
        :raises ValueError: If connection type is wrong.

        """
        line = line.strip()
        record_fields = ['conntype', 'databases', 'users']
        values = shlex.split(line, comments=False)
        # Split databases and users lists.
        values[1] = values[1].split(',')  # type: ignore
        values[2] = values[2].split(',')  # type: ignore
        try:
            hash_pos = values.index('#')
        except ValueError:
            comment = None
        else:
            values, comments = values[:hash_pos], values[hash_pos:]
            comment = ' '.join(comments[1:])

        if values[0] not in cls.CONNECTION_TYPES:
            raise ValueError("Unknown connection types %s" % values[0])
        if 'local' != values[0]:
            record_fields.append('address')
        common_values = [v for v in values if '=' not in v]
        if len(common_values) >= 6:
            record_fields.append('netmask')
        record_fields.append('method')
        base_options = list(zip(record_fields, values[:len(record_fields)]))
        auth_options = [o.split('=', 1) for o in values[len(record_fields):]]
        options = base_options + auth_options  # type: ignore
        return cls(options, comment=comment)

    conntype: Optional[str]
    databases: List[str]
    users: List[str]

    def __init__(
        self,
        values: Optional[
            Union[Iterable[Tuple[str, str]], Dict[str, Any]]
        ] = None,
        comment: Optional[str] = None,
        **kw_values: Union[str, Sequence[str]],
    ) -> None:
        """
        :param values: A dict of fields.
        :param kw_values: Fields passed as keyword.
        :param comment:  Comment at the end of the line.
        """
        values: Dict[str, Any] = dict(values or {}, **kw_values)
        if 'database' in values:
            values['databases'] = [values.pop('database')]
        if 'user' in values:
            values['users'] = [values.pop('user')]
        self.__dict__.update(values)
        self.fields = [k for k, _ in values.items()]
        self.comment = comment

    def __repr__(self) -> str:
        return '<%s %s%s>' % (
            self.__class__.__name__,
            ' '.join(self.common_values),
            '...' if self.auth_options else ''
        )

    def __str__(self) -> str:
        """Serialize a record line, without EOL."""
        # Stolen from default pg_hba.conf
        widths = [8, 16, 16, 16, 8]

        fmt = ''
        for i, field in enumerate(self.COMMON_FIELDS):
            try:
                width = widths[i]
            except IndexError:
                width = 0

            if field not in self.fields:
                fmt += ' ' * width
                continue

            if width:
                fmt += '%%(%s)-%ds ' % (field, width - 1)
            else:
                fmt += '%%(%s)s ' % (field,)
        # Serialize database and user list using property.
        values = dict(self.__dict__, databases=self.database, users=self.user)
        line = fmt.rstrip() % values

        auth_options = ['%s="%s"' % i for i in self.auth_options]
        if auth_options:
            line += ' ' + ' '.join(auth_options)

        if self.comment is not None:
            line += '  # ' + self.comment
        else:
            line = line.rstrip()

        return line

    @property
    def common_values(self) -> List[str]:
        str_fields = self.COMMON_FIELDS[:]
        # Use serialized variant.
        str_fields[1:3] = ['database', 'user']
        return [
            getattr(self, f)
            for f in str_fields
            if f in self.fields
        ]

    @property
    def auth_options(self) -> List[Tuple[str, str]]:
        return [
            (f, getattr(self, f))
            for f in self.fields
            if f not in self.COMMON_FIELDS
        ]

    @property
    def database(self) -> str:
        """Hold database column as a single value.

        Use `databases` attribute to get parsed database list. `database` is
        guaranteed to be a string.

        """
        return ','.join(self.databases)

    @property
    def user(self) -> str:
        """Hold user column as a single value.

        Use ``users`` property to get parsed user list. ``user`` is guaranteed
        to be a string.

        """
        return ','.join(self.users)

    def matches(self, **attrs: str) -> bool:
        """Tells if the current record is matching provided attributes.

        :param attrs: keyword/values pairs corresponding to one or more
            HBARecord attributes (ie. user, conntype, etc…)
        """

        # Provided attributes should be comparable to HBARecord attributes
        for k in attrs.keys():
            if k not in self.COMMON_FIELDS + ['database', 'user']:
                raise AttributeError('%s is not a valid attribute' % k)

        for k, v in attrs.items():
            if getattr(self, k, None) != v:
                return False
        return True


class HBA:
    """Represents pg_hba.conf records

    .. attribute:: lines

        List of :class:`HBARecord` and comments.

    .. attribute:: path

        Path to a file. Is automatically set when calling :meth:`parse` with a
        path to a file. :meth:`save` will write to this file if set.

    .. automethod:: __iter__
    .. automethod:: parse
    .. automethod:: save
    .. automethod:: remove
    .. automethod:: merge
    """

    lines: List[Union[HBAComment, HBARecord]]
    path: Optional[str]

    def __init__(
        self, entries: Optional[Iterable[Union[HBAComment, HBARecord]]] = None
    ) -> None:
        """HBA constructor

        :param entries: A list of HBAComment or HBARecord. Optional.
        """
        if entries and not isinstance(entries, list):
            raise ValueError('%s should be a list' % entries)
        self.lines = list(entries) if entries is not None else []
        self.path = None

    def __iter__(self) -> Iterator[HBARecord]:
        """Iterate on records, ignoring comments and blank lines."""
        for line in self.lines:
            if isinstance(line, HBARecord):
                yield line

    def parse(self, fo: Iterable[str]) -> None:
        """Parse records and comments from file object

        :param fo: An iterable returning lines
        """
        for i, line in enumerate(fo):
            stripped = line.lstrip()
            record: Union[HBARecord, HBAComment]
            if not stripped or stripped.startswith('#'):
                record = HBAComment(line.replace(os.linesep, ''))
            else:
                try:
                    record = HBARecord.parse(line)
                except Exception as e:
                    raise ParseError(1 + i, line, str(e))
            self.lines.append(record)

    def save(self, fo: Optional[Union[str, IO[str]]] = None) -> None:
        """Write records and comments in a file

        :param fo: a file-like object. Is not required if :attr:`path` is set.

        Line order is preserved. Record fields are vertically aligned to match
        the columen size of column headers from default configuration file.

        .. code::

            # TYPE  DATABASE        USER            ADDRESS                 METHOD
            local   all             all                                     trust
        """  # noqa
        with open_or_return(fo or self.path, mode='w') as fo:
            for line in self.lines:
                fo.write(str(line) + os.linesep)

    def remove(
        self,
        filter: Optional[Callable[[HBARecord], bool]] = None,
        **attrs: str,
    ) -> None:
        """Remove records matching the provided attributes.

        One can for example remove all records for which user is 'david'.

        :param filter: a function to be used as filter. It is passed the record
            to test against. If it returns True, the record is removed. It is
            kept otherwise.
        :param attrs: keyword/values pairs correspond to one or more
            HBARecord attributes (ie. user, conntype, etc...)

        Usage examples:

        .. code:: python

            hba.remove(filter=lamdba r: r.user == 'david')
            hba.remove(user='david')

        """
        if filter is not None and len(attrs.keys()):
            warnings.warn('Only filter will be taken into account')

        # Attributes list to look for must not be empty
        if filter is None and not len(attrs.keys()):
            raise ValueError('Attributes dict cannot be empty')

        filter = filter or (lambda line: line.matches(**attrs))

        self.lines = [
            line for line in self.lines
            if not (isinstance(line, HBARecord) and filter(line))
        ]

    def merge(self, other: "HBA") -> None:
        """Add new records to HBAFile or replace them if they are matching
            (ie. same conntype, database, user and address)

        :param other: HBAFile to merge into the current one.
            Lines with matching conntype, database, user and database will be
            replaced by the new one. Otherwise they will be added at the end.
            Comments from the original hba are preserved.
        """
        lines = self.lines[:]
        new_lines = other.lines[:]
        other_comments = []

        for i, line in enumerate(lines):
            if isinstance(line, HBAComment):
                continue
            for new_line in new_lines:
                if isinstance(new_line, HBAComment):
                    # preserve comments until next record
                    other_comments.append(new_line)
                else:
                    kwargs = dict()
                    for a in ['conntype', 'database', 'user', 'address']:
                        if hasattr(new_line, a):
                            kwargs[a] = getattr(new_line, a)
                    if line.matches(**kwargs):
                        # replace matched line with comments + record
                        self.lines[i:i+1] = other_comments + [new_line]  # type: ignore  # noqa: E501
                        for c in other_comments:
                            new_lines.remove(c)
                        new_lines.remove(new_line)
                        break  # found match, go to next line
                    other_comments[:] = []
        # Then add remaining new lines (not merged)
        self.lines.extend(new_lines)


def parse(file: Union[str, Iterable[str]]) -> HBA:
    """Parse a `pg_hba.conf` file.

    :param file: Either a line iterator such as a file-like object or a string
        corresponding to the path to the file to open and parse.
    :rtype: :class:`HBA`.
    """
    if isinstance(file, str):
        with open(file) as fo:
            hba = parse(fo)
            hba.path = file
    else:
        hba = HBA()
        hba.parse(file)
    return hba


if __name__ == '__main__':  # pragma: nocover
    argv = sys.argv[1:] + ['-']
    try:
        with open_or_stdin(argv[0]) as fo:
            hba = parse(fo)
        hba.save(sys.stdout)
    except Exception as e:
        print(str(e), file=sys.stderr)
        exit(1)



class ParseError(Exception):
    def __init__(self, lineno: int, line: str, message: str) -> None:
        super(ParseError, self).__init__(message)
        self.lineno = lineno
        self.line = line

    def __repr__(self) -> str:
        return '<%s at line %d: %.32s>' % (
            self.__class__.__name__, self.lineno, self.args[0])

    def __str__(self) -> str:
        return "Bad line #%s '%.32s': %s" % (
            self.lineno, self.line.strip(), self.args[0],
        )



""".. currentmodule:: pgtoolkit.service

This module supports reading, validating, editing and rendering ``pg_service``
file. See `The Connection Service File
<https://www.postgresql.org/docs/current/static/libpq-pgservice.html>`__ in
PostgreSQL documentation.

API Reference
-------------

The main entrypoint of the API is the :func:`parse` function. :func:`find`
function may be useful if you need to search for ``pg_service.conf`` files in
regular locations.

.. autofunction:: find
.. autofunction:: parse
.. autoclass:: Service
.. autoclass:: ServiceFile


Edit a service file
-------------------

.. code:: python

    from pgtoolkit.service import parse, Service

    servicefilename = 'my_service.conf'
    with open(servicefile) as fo:
        servicefile = parse(fo, source=servicefilename)

    myservice = servicefile['myservice']
    myservice.host = 'newhost'
    # Update service file
    servicefile.add(myservice)

    newservice = Service(name='newservice', host='otherhost')
    servicefile.add(newservice)

    with open(servicefile, 'w') as fo:
        servicefile.save(fo)

Shorter version using the file directly in `parse`:

.. code:: python

    servicefile = parse('my_service.conf')
    [...]
    servicefile.save()


Load a service file to connect with psycopg2
--------------------------------------------

Actually, psycopg2 already support pgservice file. This is just a showcase.

.. code:: python

    from pgtoolkit.service import find, parse
    from psycopg2 import connect

    servicefilename = find()
    with open(servicefile) as fo:
        servicefile = parse(fo, source=servicefilename)
    connection = connect(**servicefile['myservice'])


Using as a script
-----------------

:mod:`pgtoolkit.service` is usable as a CLI script. It accepts a service file
path as first argument, read it, validate it and re-render it, loosing
comments.

:class:`ServiceFile` is less strict than `libpq`. Spaces are accepted around
`=`. The output conform strictly to `libpq` parser.

.. code:: console

    $ python -m pgtoolkit.service data/pg_service.conf
    [mydb]
    host=somehost
    port=5433
    user=admin

    [my ini-style]
    host=otherhost

"""


from configparser import ConfigParser
import os
import sys
from typing import Dict, IO, Iterable, MutableMapping, Optional, Union

from ._helpers import open_or_stdin


Parameter = Union[str, int]


class Service(Dict[str, Parameter]):
    """Service definition.

    The :class:`Service` class represents a single service definition in a
    Service file. It’s actually a dictionnary of its own parameters.

    The ``name`` attributes is mapped to the section name of the service in the
    Service file.

    Each parameters can be accessed either as a dictionnary entry or as an
    attributes.

    >>> myservice = Service('myservice', {'dbname': 'mydb'}, host='myhost')
    >>> myservice.name
    'myservice'
    >>> myservice.dbname
    'mydb'
    >>> myservice['dbname']
    'mydb'
    >>> myservice.user = 'myuser'
    >>> list(sorted(myservice.items()))
    [('dbname', 'mydb'), ('host', 'myhost'), ('name', 'myservice'), ('user', 'myuser')]

    """  # noqa

    def __init__(
        self,
        name: str,
        parameters: Optional[Dict[str, Parameter]] = None,
        **extra: Parameter,
    ) -> None:
        super(Service, self).__init__()
        self.name = name
        self.update(parameters or {})
        self.update(extra)

    def __repr__(self) -> str:
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def __getattr__(self, name: str) -> Parameter:
        return self[name]

    def __setattr__(self, name: str, value: Parameter) -> None:
        self[name] = value


class ServiceFile:
    """Service file representation, parsing and rendering.

    :class:`ServiceFile` is subscriptable. You can access service using
    ``servicefile['servicename']`` syntax.

    .. automethod:: add
    .. automethod:: parse
    .. automethod:: save

    .. attribute:: path

        Path to a file. Is automatically set when calling :meth:`parse` with a
        path to a file. :meth:`save` will write to this file if set.
    """

    path: Optional[str]

    _CONVERTERS = {
        'port': int,
    }

    def __init__(self) -> None:
        self.path = None
        self.config = ConfigParser(
            comment_prefixes=('#',),
            delimiters=('=',),
        )

    def __repr__(self) -> str:
        return '<%s>' % (self.__class__.__name__)

    def __getitem__(self, key: str) -> Service:
        parameters = dict([
            (k, self._CONVERTERS.get(k, str)(v))
            for k, v in self.config.items(key)
        ])
        return Service(key, parameters)

    def __len__(self) -> int:
        return len(self.config.sections())

    def add(self, service: Service) -> None:
        """Adds a :class:`Service` object to the service file."""
        self.config.remove_section(service.name)
        self.config.add_section(service.name)
        for parameter, value in service.items():
            self.config.set(service.name, parameter, str(value))

    def parse(self, fo: Iterable[str], source: Optional[str] = None) -> None:
        """Add service from a service file.

        This method is strictly the same as :func:`parse`. It’s the method
        counterpart.
        """
        self.config.read_file(fo, source=source)

    def save(self, fo: Optional[IO[str]] = None) -> None:
        """Writes services in ``fo`` file-like object.

        :param fo: a file-like object. Is not required if :attr:`path` is set.

        .. note:: Comments are not preserved.
        """
        config = self.config

        def _write(fo: IO[str]) -> None:
            config.write(fo, space_around_delimiters=False)

        if fo:
            _write(fo)
        elif self.path:
            with open(self.path, 'w') as fo:
                _write(fo)
        else:
            raise ValueError('No file-like object nor path provided')


def guess_sysconfdir(environ: MutableMapping[str, str] = os.environ) -> str:
    fromenv = environ.get('PGSYSCONFDIR')
    if fromenv:
        candidates = [fromenv]
    else:
        candidates = [
            # From PGDG APT packages.
            '/etc/postgresql-common',
            # From PGDG RPM packages.
            '/etc/sysconfig/pgsql',
        ]

    for candidate in candidates:
        if candidate and os.path.isdir(candidate):
            return candidate
    raise Exception("Can't find sysconfdir")


def find(environ: Optional[MutableMapping[str, str]] = None) -> str:
    """Find service file.

    :param dict environ: Dict of environment variables.

    :func:`find` searches for the first candidate of ``pg_service.conf`` file
    from either environment and regular locations. :func:`find` raises an
    Exception if it fails to find a Connection service file.

    .. code:: python

        from pgtoolkit.service import find

        try:
            servicefile = find()
        except Exception as e:
            "Deal with exception."
        else:
            "Manage servicefile."

    """
    if environ is None:
        environ = os.environ

    fromenv = environ.get('PGSERVICEFILE')
    if fromenv:
        candidates = [fromenv]
    else:
        candidates = [os.path.expanduser("~/.pg_service.conf")]
        try:
            sysconfdir = guess_sysconfdir(environ)
        except Exception:
            pass
        else:
            candidates.append(os.path.join(sysconfdir, 'pg_service.conf'))

    for candidate in candidates:
        if os.path.exists(candidate):
            return candidate
    raise Exception("Can't find pg_service file.")


def parse(
    file: Union[str, Iterable[str]], source: Optional[str] = None
) -> ServiceFile:
    """Parse a service file.

    :param file: a file-object as returned by open or a string corresponding to
        the path to a file to open and parse.
    :param source: Name of the source.
    :rtype: A ``ServiceFile`` object.

    Actually it only requires as ``fo`` an iterable object yielding each lines
    of the file. You can provide ``source`` to have more precise error message.

    .. warning::

        pgtoolkit is less strict than `libpq`. `libpq` does not accepts spaces
        around equals. pgtoolkit accepts spaces but do not write them.

    """
    if isinstance(file, str):
        with open(file) as fo:
            services = parse(fo, source=source)
            services.path = file
    else:
        services = ServiceFile()
        services.parse(file, source=source)
    return services


if __name__ == '__main__':  # pragma: nocover
    argv = sys.argv[1:] + ['-']
    try:
        with open_or_stdin(argv[0]) as fo:
            services = parse(fo)
        services.save(sys.stdout)
    except Exception as e:
        print(str(e), file=sys.stderr)
        exit(1)



from setuptools import setup


metadatas = dict(
    name='pgtoolkit',
    description='Postgres Support from Python',
    version='0.8.2',
    author='Dalibo',
    author_email='contact@dalibo.com',
    url='https://github.com/dalibo/pgtoolkit',
    license='PostgreSQL',
)


if __name__ == '__main__':
    setup(
        packages=['pgtoolkit'],
        package_data={'pgtoolkit': ['py.typed']},
        python_requires=">=3.6",
        **metadatas
    )



"""\
.. currentmodule:: pgtoolkit.conf

This module implements ``postgresql.conf`` file format. This is the same format
for ``recovery.conf``. The main entry point of the API is :func:`parse`. The
module can be used as a CLI script.


API Reference
-------------

.. autofunction:: parse
.. autoclass:: Configuration


Using as a CLI Script
---------------------

You can use this module to dump a configuration file as JSON object

.. code:: console

    $ python -m pgtoolkit.conf postgresql.conf | jq .
    {
      "lc_monetary": "fr_FR.UTF8",
      "datestyle": "iso, dmy",
      "log_rotation_age": "1d",
      "log_min_duration_statement": "3s",
      "log_lock_waits": true,
      "log_min_messages": "notice",
      "log_directory": "log",
      "port": 5432,
      "log_truncate_on_rotation": true,
      "log_rotation_size": 0
    }
    $

"""


import json
from ast import literal_eval
from collections import OrderedDict
import re
import sys
from datetime import timedelta
from typing import Dict, IO, Iterable, List, Optional, Union

from ._helpers import JSONDateEncoder
from ._helpers import open_or_stdin
from ._helpers import open_or_return


def parse(fo: Union[str, IO[str]]) -> "Configuration":
    """Parse a configuration file.

    The parser tries to return Python object corresponding to value, based on
    some heuristics. booleans, octal number, decimal integers and floating
    point numbers are parsed. Multiplier units like kB or MB are applyied and
    you get an int. Interval value like ``3s`` are returned as
    :class:`datetime.timedelta`.

    In case of doubt, the value is kept as a string. It's up to you to enforce
    format.

    :param fo: A line iterator such as a file-like object or a path.
    :returns: A :class:`Configuration` containing parsed configuration.

    """
    conf = Configuration()
    with open_or_return(fo) as fo:
        conf.parse(fo)
        conf.path = getattr(fo, 'name', None)
    return conf


MEMORY_MULTIPLIERS = {
    'kB': 1024,
    'MB': 1024 * 1024,
    'GB': 1024 * 1024 * 1024,
    'TB': 1024 * 1024 * 1024 * 1024,
}
_memory_re = re.compile(r'^\s*(?P<number>\d+)\s*(?P<unit>[kMGT]B)\s*$')
TIMEDELTA_ARGNAME = {
    'ms': 'milliseconds',
    's': 'seconds',
    'min': 'minutes',
    'h': 'hours',
    'd': 'days',
}
_timedelta_re = re.compile(r'^\s*(?P<number>\d+)\s*(?P<unit>ms|s|min|h|d)\s*$')


Value = Union[str, bool, float, int, timedelta]


def parse_value(raw: str) -> Value:
    # Ref.
    # https://www.postgresql.org/docs/current/static/config-setting.html#CONFIG-SETTING-NAMES-VALUES

    if raw.startswith("'"):
        try:
            raw = literal_eval(raw)
        except SyntaxError as e:
            raise ValueError(str(e))

    if raw.startswith('0'):
        try:
            return int(raw, base=8)
        except ValueError:
            pass

    m = _memory_re.match(raw)
    if m:
        unit = m.group('unit')
        mul = MEMORY_MULTIPLIERS[unit]
        return int(m.group('number')) * mul

    m = _timedelta_re.match(raw)
    if m:
        unit = m.group('unit')
        arg = TIMEDELTA_ARGNAME[unit]
        kwargs = {arg: int(m.group('number'))}
        return timedelta(**kwargs)

    elif raw in ('true', 'yes', 'on'):
        return True
    elif raw in ('false', 'no', 'off'):
        return False
    else:
        try:
            return int(raw)
        except ValueError:
            try:
                return float(raw)
            except ValueError:
                return raw


class Entry:
    # Holds the parsed representation of a configuration entry line.
    #
    # This includes the comment.

    def __init__(
        self,
        name: str,
        value: Value,
        comment: Optional[str] = None,
        raw_line: Optional[str] = None,
    ) -> None:
        self.name = name
        self.value = value
        self.comment = comment
        # Store the raw_line to track the position in the list of lines.
        self.raw_line = raw_line

    def __repr__(self) -> str:
        return '<%s %s=%s>' % (self.__class__.__name__, self.name, self.value)

    _minute = 60
    _hour = 60 * _minute
    _day = 24 * _hour

    _timedelta_unit_map = [
        ('d', _day),
        ('h', _hour),
        # The space before 'min' is intentionnal. I find '1 min' more readable
        # than '1min'.
        (' min', _minute),
        ('s', 1),
    ]

    def serialize(self) -> Union[int, str]:
        # This is the reverse of parse_value.
        value = self.value
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, int):
            for unit in None, 'kB', 'MB', 'GB', 'TB':
                if value % 1024:
                    break
                value = value // 1024
            if unit:
                value = "'%s %s'" % (value, unit)
        elif isinstance(value, str):
            if ' ' in value or "'" in value:
                value = "'%s'" % value.replace("'", r"\'")
        elif isinstance(value, timedelta):
            seconds = value.days * self._day + value.seconds
            if value.microseconds:
                unit = ' ms'
                value = seconds * 1000 + value.microseconds // 1000
            else:
                for unit, mod in self._timedelta_unit_map:
                    if seconds % mod:
                        continue
                    value = seconds // mod
                    break
            value = "'%s%s'" % (value, unit)
        else:
            value = str(value)
        return value

    def __str__(self) -> str:
        line = '%(name)s = %(value)s' % dict(
            name=self.name, value=self.serialize())
        if self.comment:
            line += '  # ' + self.comment
        return line


class Configuration:
    r"""Holds a parsed configuration.

    You can access parameter using attribute or dictionnary syntax.

    >>> conf = parse(['port=5432\n', 'pg_stat_statement.min_duration = 3s\n'])
    >>> conf.port
    5432
    >>> conf.port = 5433
    >>> conf.port
    5433
    >>> conf['port'] = 5434
    >>> conf.port
    5434
    >>> conf['pg_stat_statement.min_duration'].total_seconds()
    3.0

    .. attribute:: path

        Path to a file. Automatically set when calling :func:`parse` with a path
        to a file. This is default target for :meth:`save`.

    .. automethod:: save

    """  # noqa
    lines: List[str]
    entries: Dict[str, Entry]
    path: Optional[str]

    _parameter_re = re.compile(
        r'^(?P<name>[a-z_.]+)(?: +(?!=)| *= *)(?P<value>.*?)'
        '[\\s\t]*'
        r'(?P<comment>#.*)?$'
    )

    # Internally, lines property contains an updated list of all comments and
    # entries serialized. When adding a setting or updating an existing one,
    # the serialized line is updated accordingly. This allows to keep comments
    # and serialize only what's needed. Other lines are just written as-is.

    def __init__(self) -> None:
        self.__dict__.update(dict(
            lines=[],
            entries=OrderedDict(),
            path=None,
        ))

    def parse(self, fo: Iterable[str]) -> None:
        for raw_line in fo:
            self.lines.append(raw_line)
            line = raw_line.strip()
            if not line or line.startswith('#'):
                continue

            m = self._parameter_re.match(line)
            if not m:
                raise ValueError("Bad line: %r." % raw_line)
            kwargs = m.groupdict()
            value = parse_value(kwargs.pop('value'))
            entry = Entry(value=value, raw_line=raw_line, **kwargs)
            self.entries[entry.name] = entry

    def __getattr__(self, name: str) -> Value:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Value) -> None:
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self[name] = value

    def __getitem__(self, key: str) -> Value:
        return self.entries[key].value

    def __setitem__(self, key: str, value: Value) -> None:
        if key in self.entries:
            e = self.entries[key]
            e.value = value
            # Update serialized entry.
            assert e.raw_line is not None
            old_line = e.raw_line
            e.raw_line = str(e) + '\n'
            lineno = self.lines.index(old_line)
            self.lines[lineno:lineno+1] = [e.raw_line]
        else:
            self.entries[key] = e = Entry(name=key, value=value)
            # Append serialized line.
            e.raw_line = str(e) + '\n'
            self.lines.append(e.raw_line)

    def as_dict(self) -> Dict[str, Value]:
        return dict([(k, v.value) for k, v in self.entries.items()])

    def save(self, fo: Optional[Union[str, IO[str]]] = None) -> None:
        """Write configuration to a file.

        Configuration entries order and comments are preserved.

        :param fo: A path or file-like object. Required if :attr:`path` is
            None.

        """
        with open_or_return(fo or self.path, mode='w') as fo:
            for line in self.lines:
                fo.write(line)


def _main(argv: List[str]) -> int:  # pragma: nocover
    argv = argv or ['-']
    try:
        with open_or_stdin(argv[0]) as fo:
            conf = parse(fo)
        print(json.dumps(conf.as_dict(), cls=JSONDateEncoder, indent=2))
        return 0
    except Exception as e:
        print(str(e), file=sys.stderr)
        return 1


if __name__ == '__main__':  # pragma: nocover
    exit(_main(sys.argv[1:]))



r""".. currentmodule:: pgtoolkit.pgpass

This module provides support for `.pgpass` file format. Here are some
highlights :

 - Supports ``:`` and ``\`` escape.
 - Sorts entry by precision (even if commented).
 - Preserves comments order when sorting.

See `The Password File
<https://www.postgresql.org/docs/current/static/libpq-pgpass.html>`__ section
in PostgreSQL documentation.

.. autofunction:: parse
.. autoclass:: PassEntry
.. autoclass:: PassComment
.. autoclass:: PassFile


Editing a .pgpass file
----------------------

.. code:: python

    with open('.pgpass') as fo:
        pgpass = parse(fo)
    pgpass.lines.append(PassEntry(username='toto', password='confidentiel'))
    pgpass.sort()
    with open('.pgpass', 'w') as fo:
        pgpass.save(fo)

Shorter version using the file directly in `parse`:

.. code:: python

    pgpass = parse('.pgpass')
    pgpass.lines.append(PassEntry(username='toto', password='confidentiel'))
    pgpass.sort()
    pgpass.save()

Using as a script
-----------------

You can call :mod:`pgtoolkit.pgpass` module as a CLI script. It accepts a file
path as first argument, read it, validate it, sort it and output it in stdout.


.. code:: console

   $ python -m pgtoolkit.pgpass ~/.pgpass
   more:5432:precise:entry:0revea\\ed
   #disabled:5432:*:entry:0secret

   # Multiline
   # comment.
   other:5432:*:username:0unveiled
   *:*:*:postgres:c0nfident\:el

"""  # noqa

import os
import sys
from typing import (
    Callable,
    IO,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
import warnings

from .errors import ParseError
from ._helpers import open_or_stdin


def unescape(s: str, delim: str) -> str:
    return s.replace('\\' + delim, delim).replace('\\\\', '\\')


def escapedsplit(s: str, delim: str) -> Iterator[str]:
    if len(delim) != 1:
        raise ValueError('Invalid delimiter: ' + delim)

    ln = len(s)
    escaped = False
    i = 0
    j = 0

    while j < ln:
        if s[j] == '\\':
            escaped = not escaped
        elif s[j] == delim:
            if not escaped:
                yield unescape(s[i:j], delim)
                i = j + 1
                escaped = False
        j += 1
    yield unescape(s[i:j], delim)


class PassComment(str):
    """A .pgpass comment, including spaces and ``#``.

    It's a child of ``str``.

    >>> comm = PassComment("# my comment")
    >>> comm.comment
    'my comment'

    .. automethod:: matches

    .. attribute:: comment

        The actual message of the comment. Surrounding whitespaces stripped.

    """
    def __repr__(self) -> str:
        return '<%s %.32s>' % (self.__class__.__name__, self)

    def __lt__(self, other: str) -> bool:
        if isinstance(other, PassEntry):
            try:
                return self.entry < other
            except ValueError:
                pass
        return False

    @property
    def comment(self) -> str:
        return self.lstrip('#').strip()

    @property
    def entry(self) -> "PassEntry":
        if not hasattr(self, '_entry'):
            self._entry = PassEntry.parse(self.comment)
        return self._entry

    def matches(self, **attrs: Union[int, str]) -> bool:
        """In case of a commented entry, tells if it is matching provided
        attributes. Returns False otherwise.

        :param attrs: keyword/values pairs correspond to one or more
            PassEntry attributes (ie. hostname, port, etc...)
        """
        try:
            return self.entry.matches(**attrs)
        except ValueError:
            return False


class PassEntry:
    """Holds a .pgpass entry.

    .. automethod:: parse
    .. automethod:: matches

    .. attribute:: hostname

       Server hostname, the first field.

    .. attribute:: port

       Server port, the second field.

    .. attribute:: database

       Database, the third field.

    .. attribute:: username

       Username, the fourth field.

    .. attribute:: password

       Password, the fifth field.

    :class:`PassEntry` object is sortable. A :class:`PassEntry` object is lower
    than another if it is more specific. The more an entry has wildcard, the
    less it is specific.

    """

    @classmethod
    def parse(cls, line: str) -> "PassEntry":
        """ Parse a single line.

        :param line: string containing a serialized .pgpass entry.
        :return: :class:`PassEntry` object holding entry data.
        :raises ValueError: on invalid line.
        """
        fields = list(escapedsplit(line.strip(), ':'))
        if len(fields) != 5:
            raise ValueError("Invalid line.")
        if fields[1] != '*':
            fields[1] = int(fields[1])  # type: ignore
        return cls(*fields)

    def __init__(
        self,
        hostname: str,
        port: Union[int, str],
        database: str,
        username: str,
        password: str
    ) -> None:
        self.hostname = hostname
        self.port = port
        self.database = database
        self.username = username
        self.password = password

    def __eq__(self, other: object) -> bool:
        if isinstance(other, PassComment):
            try:
                other = other.entry
            except ValueError:
                return False
        if isinstance(other, PassEntry):
            return self.as_tuple()[:-1] == other.as_tuple()[:-1]
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.as_tuple()[:-1])

    def __lt__(self, other: Union[PassComment, "PassEntry"]) -> bool:
        if isinstance(other, PassComment):
            try:
                other = other.entry
            except ValueError:
                return False
        if isinstance(other, PassEntry):
            return self.sort_key() < other.sort_key()
        return NotImplemented

    def __repr__(self) -> str:
        return '<%s %s@%s:%s/%s>' % (
            self.__class__.__name__,
            self.username, self.hostname, self.port, self.database,
        )

    def __str__(self) -> str:
        return ':'.join([
            str(x).replace('\\', r'\\').replace(':', r'\:')
            for x in self.as_tuple()
        ])

    def as_tuple(self) -> Tuple[str, Union[int, str], str, str, str]:
        return (
            self.hostname,
            self.port,
            self.database,
            self.username,
            self.password,
        )

    def sort_key(self) -> Tuple[int, str, Union[int, str], str, str]:
        tpl = self.as_tuple()[:-1]
        # Compute precision from * occurences.
        precision = len([x for x in tpl if x == '*'])
        # More specific entries comes first.
        return (precision,) + tuple(chr(0xFF) if x == '*' else x for x in tpl)  # type: ignore  # noqa: E501

    def matches(self, **attrs: Union[int, str]) -> bool:
        """Tells if the current entry is matching provided attributes.

        :param attrs: keyword/values pairs correspond to one or more
            PassEntry attributes (ie. hostname, port, etc...)
        """

        # Provided attributes should be comparable to PassEntry attributes
        expected_attributes = self.__dict__.keys()
        for k in attrs.keys():
            if k not in expected_attributes:
                raise AttributeError('%s is not a valid attribute' % k)

        for k, v in attrs.items():
            if getattr(self, k) != v:
                return False
        return True


class PassFile:
    """Holds .pgpass file entries and comments.

    .. automethod:: parse
    .. automethod:: __iter__
    .. automethod:: sort
    .. automethod:: save
    .. automethod:: remove

    .. attribute:: lines

        List of either :class:`PassEntry` or :class:`PassFile`. You can add
        lines by appending :class:`PassEntry` or :class:`PassFile` instances to
        this list.

    .. attribute:: path

        Path to a file. Is automatically set when calling :meth:`parse` with a
        path to a file. :meth:`save` will write to this file if set.

    """

    lines: List[Union[PassComment, PassEntry]]
    path: Optional[str] = None

    def __init__(
        self,
        entries: Optional[List[Union[PassComment, PassEntry]]] = None
    ) -> None:
        """PassFile constructor.

        :param entries: A list of PassEntry or PassComment. Optional.
        """
        if entries and not isinstance(entries, list):
            raise ValueError('%s should be a list' % entries)
        self.lines = entries or []
        self.path = None

    def __iter__(self) -> Iterator[PassEntry]:
        """Iterate entries

        Yield :class:`PassEntry` instance from parsed file, ignoring comments.
        """
        for line in self.lines:
            if isinstance(line, PassEntry):
                yield line

    def parse(self, fo: Iterable[str]) -> None:
        """Parse lines

        :param fo: A line iterator such as a file-like object.

        Raises ``ParseError`` if a bad line is found.
        """
        entry: Union[PassComment, PassEntry]
        for i, line in enumerate(fo):
            stripped = line.lstrip()
            if not stripped or stripped.startswith('#'):
                entry = PassComment(line.replace(os.linesep, ''))
            else:
                try:
                    entry = PassEntry.parse(line)
                except Exception as e:
                    raise ParseError(1 + i, line, str(e))
            self.lines.append(entry)

    def sort(self) -> None:
        """Sort entries preserving comments.

        libpq use the first entry from .pgpass matching connexion informations.
        Thus, less specific entries should be last in the file. This is the
        purpose of :func:`sort` method.

        About comments. Comments are supposed to bear with the entrie
        **below**. Thus comments block are sorted according to the first entry
        below.

        Commented entries are sorted like entries, not like comment.
        """
        # Sort but preserve comments above entries.
        entries = []
        comments = []
        for line in self.lines:
            if isinstance(line, PassComment):
                try:
                    line.entry
                except ValueError:
                    comments.append(line)
                    continue

            entries.append((line, comments))
            comments = []

        entries.sort()
        self.lines[:] = []
        for entry, comments in entries:
            self.lines.extend(comments)
            self.lines.append(entry)

    def save(self, fo: Optional[IO[str]] = None) -> None:
        """Save entries and comment in a file.

        :param fo: a file-like object. Is not required if :attr:`path` is set.
        """
        def _write(fo: IO[str], lines: Iterable[object]) -> None:
            for line in lines:
                fo.write(str(line) + os.linesep)

        if fo:
            _write(fo, self.lines)
        elif self.path:
            with open(self.path, 'w') as fo:
                _write(fo, self.lines)
        else:
            raise ValueError('No file-like object nor path provided')

    def remove(
        self,
        filter: Optional[
            Callable[[Union[PassComment, PassEntry, str]], bool]
        ] = None,
        **attrs: Union[int, str],
    ) -> None:
        """Remove entries matching the provided attributes.

        One can for example remove all entries for which port is 5433.

        Note: commented entries matching will also be removed.

        :param filter: a function to be used as filter. It is passed the line
             to test against. If it returns True, the line is removed. It is
             kept otherwise.
        :param attrs: keyword/values pairs correspond to one or more
            PassEntry attributes (ie. hostname, port, etc...)

        Usage examples:

        .. code:: python

            pgpass.remove(port=5432)
            pgpass.remove(filter=lambda r: r.port != 5432)
        """
        if filter is not None and len(attrs):
            warnings.warn('Only filter will be taken into account')

        # Attributes list to look for must not be empty
        if filter is None and not len(attrs.keys()):
            raise ValueError('Attributes dict cannot be empty')

        if filter is not None:
            # Silently handle the case when line is a PassComment
            def filter_(line: Union[PassComment, PassEntry]) -> bool:
                if isinstance(line, PassComment):
                    try:
                        return filter(line.entry)  # type: ignore
                    except ValueError:
                        return False
                else:
                    return filter(line)  # type: ignore
        else:
            def filter_(line: Union[PassComment, PassEntry]) -> bool:
                return line.matches(**attrs)

        self.lines = [
            line for line in self.lines if not filter_(line)
        ]


def parse(file: Union[str, IO[str]]) -> PassFile:
    """Parses a .pgpass file.

    :param file: Either a line iterator such as a file-like object or a string
        corresponding to the path to the file to open and parse.
    :rtype: :class:`PassFile`
    """
    if isinstance(file, str):
        with open(file) as fo:
            pgpass = parse(fo)
            pgpass.path = file
    else:
        pgpass = PassFile()
        pgpass.parse(file)
    return pgpass


if __name__ == '__main__':  # pragma: nocover
    argv = sys.argv[1:] + ['-']
    try:
        with open_or_stdin(argv[0]) as fo:
            pgpass = parse(fo)
        pgpass.sort()
        pgpass.save(sys.stdout)
    except Exception as e:
        print(str(e), file=sys.stderr)
        exit(1)
